/*
   .X model loading Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to load a .x model file into our Direct3D scene.
   Just like many other things in DirectX instead of us writing a whole lot of code
   to do something, DirectX can handle everything for us with built in member
   functions.  In fact all we really need to load in a model is to call...
   D3DXLoadMeshFromX() and send in the right values for the parameters and DirectX
   will take care of everything else.  Our model in this demo is a simple square.
   I did it like this so you can view the .x model file and see how everything works
   more easily in notepad.
*/


// Demo: .X model loading.


#define WIN32_LEAN_AND_MEAN
#define VC_LEANMEAN

#include<d3dx9.h>


// Function Prototypes...
bool InitializeDirect3D(HWND hwnd, bool fullscreen);
void RenderScene();
void SetRotation();
void SetView();
void SetPerpective();
void ShutdownDirect3D();


// Here are our Direct3D objects.
LPDIRECT3D9 Direct3D_Object = NULL;
LPDIRECT3DDEVICE9 D3D_Device = NULL;

LPD3DXBUFFER MaterialBuffer = NULL;       // Material buffer.
DWORD numMaterials;                       // Total number of materials.
LPD3DXMESH Model = NULL;                  // Mesh object.
D3DMATERIAL9* MaterialList = NULL;        // List of materials for the model.
LPDIRECT3DTEXTURE9* TextureList = NULL;   // List of texture for the model.

// This is the rotation angle.
float angle = 0.0f;


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch(message)
       {
		   case WM_KEYUP:
            // If the user presses the escape key then exit the application.
            if(wParam == VK_ESCAPE)
               PostQuitMessage(0);
            break;

         case WM_DESTROY:
         case WM_CLOSE:
            PostQuitMessage(0);
            break;

            default:
               break;
       }

    return DefWindowProc( hwnd, message, wParam, lParam );
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                           // Message object.
   HWND hwnd;                                         // Handle to the window.
   WNDCLASSEX windowClass;                            // Window class object.
   bool done = false;                                 // true then we should exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);           // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;       // style of the window.
   windowClass.lpfnWndProc = WndProc;                 // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                        // Extra class information.
   windowClass.cbWndExtra = 0;                        // Extra window information.
   windowClass.hInstance = hInstance;                 // Handle of application Instance.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);// Window Icon.
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW); // Load mouse cursor.
   windowClass.hbrBackground = NULL;                  // Background color.
   windowClass.lpszMenuName = NULL;                   // Menu.
   windowClass.lpszClassName = "UGPClass";            // Name of the window class.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);// Minimized window icon.

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // After your class has been registered then you are ready to create your window.
   hwnd = CreateWindowEx(NULL,                                       // The extended style.
                         "UGPClass",                                 // Window class.
                         "X model loading - by The Programming Ace", // Window name.
                          WS_OVERLAPPEDWINDOW | WS_VISIBLE |         // Window style.
                          WS_SYSMENU |WS_CLIPCHILDREN |              // Window style.
                          WS_CLIPSIBLINGS,                           // Window style.
                          100, 100,                                  // X, Y coords.
                          640, 480,                                  // Window size.
                          NULL,                                      // Handle to parent window.
                          NULL,                                      // Menu.
                          hInstance,                                 // Handle to app instance.
                          NULL);                                     // Pointer to window.

   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW);    // Show the window.
   UpdateWindow(hwnd);           // Update its display.

   done = false;                 // false = run program, true means stop.

   // Initialize Direct3D.  If fails then we don't want to run the application.
   if(!InitializeDirect3D(hwnd, false)) done = true;

   // Application loop.
   while(!done)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     done = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
            }
      }

    // Here we shutdown Direct 3D.
	ShutdownDirect3D();

   // Here we unregister the window class with the OS.
   UnregisterClass("UGPClass", windowClass.hInstance);

   return (int)msg.wParam;
}


bool InitializeDirect3D(HWND hwnd, bool fullscreen)
{
   // This object will allow us to set the display mode of the screen.
   D3DDISPLAYMODE DisplayMode;

   // This will allow us to set the parameters of the screen.
   D3DPRESENT_PARAMETERS Present_Parameters;

   // This is used to get the capabilities of the hardware.
   D3DCAPS9 D3DCaps;

   // It is always a good idea to clear out memory in object although not necessary.
   ZeroMemory(&Present_Parameters, sizeof(Present_Parameters));

   // Create the Direct3D object to get everything started.
   Direct3D_Object = Direct3DCreate9(D3D_SDK_VERSION);

   // Error checking.  Make sure that it was successful.
   if(Direct3D_Object == NULL)
      {
         MessageBox(NULL, "Error, couldn't initialize DirectX!?!",
                    "Error!", MB_OK);
         return false;
      }

   // This function will get the display mode of the device and place it in DisplayMode.
   if(FAILED(Direct3D_Object->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &DisplayMode)))
      {
         MessageBox(NULL, "Error setting the display mode.", "Error!", MB_OK);
         return false;
      }

   // Get the capabilities of the hardware.
	if(FAILED(Direct3D_Object->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &D3DCaps)))
		return false;

   // Test which is supported, hardware or software vertex processing.
	DWORD VertexProcessing = 0;

	if(D3DCaps.VertexProcessingCaps != 0)
		VertexProcessing |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else
		VertexProcessing |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;

   // Here we are setting the applications parameters...
   if(fullscreen)
      {
         Present_Parameters.Windowed = FALSE;               // Window mode (fullscreen).
         Present_Parameters.BackBufferWidth = DisplayMode.Width;
         Present_Parameters.BackBufferHeight = DisplayMode.Height;
      }
   else
      Present_Parameters.Windowed = TRUE;                   // Window mode (not fullscreen).
   Present_Parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;   // Dealing with animation (see doc).
   Present_Parameters.BackBufferFormat = DisplayMode.Format;// Render to the area of the screen.
   Present_Parameters.BackBufferCount = 1;                  // Number of back buffers.
   Present_Parameters.EnableAutoDepthStencil = TRUE;        // Check documentation.
   Present_Parameters.AutoDepthStencilFormat = D3DFMT_D16;  // Check documentation.

   // Now we must create the rendering device.
   if(FAILED(Direct3D_Object->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL,
                                           hwnd, VertexProcessing,
                                           &Present_Parameters, &D3D_Device)))
      {
         MessageBox(NULL, "CreateDevice() failed!  Make sure you have DirectX 9.",
                    "Error!", MB_OK);
         return false;
      }

   // One last check to be sure.
   if(D3D_Device == NULL)
      {
         MessageBox(NULL, "D3D_Device is equal to NULL!?!", "Error!", MB_OK);
         return false;
      }

   D3D_Device->SetRenderState(D3DRS_LIGHTING, FALSE);
	D3D_Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
   D3D_Device->SetRenderState(D3DRS_ZENABLE, TRUE);

   // Use the D3DXLoadMeshFromX() to load a .x model file into the mesh object.  The
   // first parameter is the file name, the second is options, the third is the
   // device, the fourth is a pointer that contains adjacent data (check documentation),
   // the fiths is a list of materials, the sixth is a list of effect instances,
   // the seventh is the number of materials, while the last is the mesh object itself.
   if(FAILED(D3DXLoadMeshFromX("Model.x", D3DXMESH_SYSTEMMEM, D3D_Device,
                     NULL, &MaterialBuffer, NULL, &numMaterials, &Model)))
      {
         MessageBox(NULL, "Error loading Model.x model file...", "Error!", MB_OK);
         return false;
      }

   // Allocate the list of materials based on the total number of materials loaded from the x file.
   MaterialList = new D3DMATERIAL9[numMaterials];
   TextureList  = new LPDIRECT3DTEXTURE9[numMaterials];

   // Get a pointer to the buffer that was sent to D3DXLoadMeshFromX().
   D3DXMATERIAL* matMaterials = (D3DXMATERIAL*)MaterialBuffer->GetBufferPointer();

   // Loop through and load each material and any textures the model uses.
   for(DWORD i = 0; i < numMaterials; i++)
      {
         // Copy the materials from the buffer into our list.
         MaterialList[i] = matMaterials[i].MatD3D;
         MaterialList[i].Ambient = MaterialList[i].Diffuse;

         // Load the textures into the list.  If this material does not have a texture the
         // index in this list will be NULL so no worries.
         if(FAILED(D3DXCreateTextureFromFile(D3D_Device, matMaterials[i].pTextureFilename, 
                                             &TextureList[i])))
            {
               TextureList[i] = NULL;
            }
      }

   // Set the texture min/mag info.
   D3D_Device->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	D3D_Device->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

   // Set the window perpective.
   SetPerpective();

   return true;
}


void SetPerpective()
{
   // This matrix will do nothing more but tell Direct3D how far back it can see
   // and how close it can see.
   D3DXMATRIXA16 Projection;

   // Set the perspective information.
   D3DXMatrixPerspectiveFovLH(&Projection, 45.0f, 640/480, 0.1f, 500.0f);

   // Set the projection.
   D3D_Device->SetTransform(D3DTS_PROJECTION, &Projection);
}


void SetRotation()
{
   // This will hold the world matrix.
   D3DXMATRIXA16 World;

   // Set the rotation value to the matrix.
   D3DXMatrixRotationY(&World, angle);

   // Set the matrix.
   D3D_Device->SetTransform(D3DTS_WORLD, &World);

   // Add 0.01 to angle to rotate the triangle even more the next frame.
   angle += 0.01f;

   if(angle >= 360)
      angle = 0.0f;
}


void SetView()
{
   // This will hold the view matrix.
   D3DXMATRIXA16 View;

   D3DXVECTOR3 Eye(0.0f, 0.0f, -5.0f);     // Position of view.
   D3DXVECTOR3 LookAt(0.0f, 0.0f, 0.0f);   // Point view is looking at.
   D3DXVECTOR3 Up(0.0f, 1.0f, 0.0f);       // Which way is up.

   // Set all the values this matrix will need to know (position, look at point, etc).
   D3DXMatrixLookAtLH(&View, &Eye, &LookAt, &Up);

   // Move the "view" to the position and values we set above.
   D3D_Device->SetTransform(D3DTS_VIEW, &View);
}


void RenderScene()
{
   // Clear the screen to black.
   D3D_Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                     D3DCOLOR_XRGB(0,0,0), 1.0f, 0);
    
   // Tell Direct 3D we are now going to start drawing.
   D3D_Device->BeginScene();

      // Set the world matrix so we can rotate and animate our triangle.
      SetRotation();

      // Set the view matrix (view position).
      SetView();

      // Loop through all materials and draw each part of the model.
      // Since there is only one part (a square) in this demo's X model file this will
      // only happen once.
      for(DWORD i = 0; i < numMaterials; i++)
         {
            D3D_Device->SetMaterial(&MaterialList[i]);
            D3D_Device->SetTexture(0, TextureList[i]);

            Model->DrawSubset(i);
         }

   // Done drawing.
   D3D_Device->EndScene();
    
   // Swap the old frame with the new one.
   D3D_Device->Present(NULL, NULL, NULL, NULL);
}


void ShutdownDirect3D()
{
   // Here we release the Direct3D objects.
   if(D3D_Device != NULL)
      {
         D3D_Device->Release();
         D3D_Device = NULL;
      }

   if(Direct3D_Object != NULL)
      {
         Direct3D_Object->Release();
         Direct3D_Object = NULL;
      }
   
   for(DWORD i = 0; i < numMaterials; i++)
      {
         if(TextureList[i] != NULL)
            {
               TextureList[i]->Release();
               TextureList[i] = NULL;
            }
      }

   if(MaterialList != NULL)
      {
         delete[] MaterialList;
         MaterialList = NULL;
      }

   if(TextureList != NULL)
      {
         delete[] TextureList;
         TextureList = NULL;
      }

   if(MaterialBuffer != NULL)
      {
         MaterialBuffer->Release();
         MaterialBuffer = NULL;
      }

   if(Model != NULL)
      {
         Model->Release();
         Model = NULL;
      }
}


// Recap:

// I know some of you are probably staring at the screen wondering if that is all there is to
// loading a .x model and at a basic level it is.  I suggest you try out a few different .x models
// out to see the results.  If you want an explanation on the .x file format I suggest you check the
// documentation and/or check out the Model.x file provided.  Just about all of the work is
// done by DirectX.


// Copyright September 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com